/*  
  prerex.c  -- interactive editor of prerequisite-chart descriptions
  Copyright (c) 2005-09  R. D. Tennent   
  School of Computing, Queen's University, rdt@cs.queensu.ca 

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*/

/* 

The key functions in this program are as follows:

  analyze_tex_file():  reads a TeX file and builds an internal representation 
    of the boxes and arrows for the chart in linked lists accessed
    through global pointer variables first_box and first_arrow, resp.
    Non-chart elements of the TeX file are preserved in temporary
    files pretext, posttext, and comments. 

  regenerate_tex_file():  re-constructs a TeX file from the temporary files
    pretext, posttext and comments, and the linked lists of nodes and arrows;
    merge_lists() implements the output-ordering strategy.

  analyze_user_command(): processes editing and other commands entered at the 
    interactive prompt. The current stacks of "cuts" and "deletions" are
    accessed through pointer variables first_cut and first_deletion.

The Boolean variables grid and reprocess record whether a background
coordinate grid should be displayed, and whether the tex file should be
reprocessed after an editing operation.

Changelog:

Version 5.4, 2009-11-12

  Must re-open tex_file to regenerate it (!)

Version 5.3, 2009-10-30

  Use ps2pdf without attempting rotation correction when using latex,
    but, to allow for a script that tests for "Orientation: Landscape" in the ps file,
    if the environment variable PS2PDF is defined, its value is used in lieu of ps2pdf.

  Offer to quit if the user won't provide a file name.

  Allow for PAGER == most.

  Code clean-up, especially localizing many globals.

Version 5.2, 2009-10-19

  Corrected coding for new file name in open_tex_file.

Version 5.1, 2009-10-01

  Final processing no longer automatically uses LaTeX.

Version 5.0, 2009-09-15

  Replaced filename by basefilename and chartfilename to allow more than
  one document to share an included chart file and allow more than one
  chart to be included in a document.  The *file* command has been removed.

Version 4.3, 2009-09-06

  Replaced strlcpy and strlcpy by append (with offset).
  Introduced triv_type to avoid a "pedantic" error.
  Corrected an incorrect sizeof argument in open_tex_file.
  Call regenerate_and_process after a partial raise.

Version 4.2, 2008-07-17

  Use the /Orientation directive to gs (ghostscript) to ensure proper orientation
  when slanted fonts are used.

Version 4.1, 2008-05-16

  Implement shift/raise of specified diagram elements.
  Eliminate unnecessary dependence on libtermcap.

Version 4.0.1, 2008-02-28

  Corrected txt field of struct text to have size LINE_LEN+1

Version 4.0, 2008-02-13

  Test for curvatures > 100.
  Implement \text command.

Version 3.8.1, 2007-03-17

  Sleep for KILL_WAIT seconds before terminating the pdf viewer.

Version I.8, 2007-02-08

  Semicolon appended to most user commands suppresses regeneration and reprocessing.
  ^C saves to file.tex. Use quit; to exit without saving.
  Final processing uses latex -> dvips -> ps2pdf to produce a smaller final pdf.
  Attempts to display the pdf during editing using value of environment variable 
    PDFVIEWER. 

Version 3.7, 2006-12-27

  latex->dvips->ps2pdf processing restored as an option.
  Warns about quitting with unpasted cuts.

Version 3.6, 2006-11-29

  Corrects two minor logical errors in code for arrow deletion and box editing. 
  Generates progress messages when analyzing/saving/processing.
  Re-loads after every shell escape, so we can discard the Edit command.
  Applies shift/raise to cut and deleted nodes.
  Processes tex files using pdflatex instead of latex -> dvips -> ps2pdf.

Version 3.5, 2006-11-20
  
  simpler and more flexible implementation of cut/paste, delete/undelete
  generate undelete messages
  automatic re-generation and processing after editing operation

Version 3.4, 2006-11-15

  deal with files that are (initially) read-only
  deal with attempts to cut arrows

Version 3.3, 2006-11-07

  preserve comment lines within the chart environment
  correct option-handling bug introduced in 3.0

Version 3.2, 2006-11-03
Version 3.1, 2006-10-29

  introduce Edit command
  restore write-access to TeX file during Edit or shell-command execution
  try to display error message in log file if LaTeX processing fails

Version 3.0, 2006-10-19

  replaced calls to make file.pdf by successive calls to latex, dvips, ps2pdf
  removed all references to pdf viewer, user must start one him/herself
  replaced strncpy/cat by strlcpy/cat (my implementation of the OpenBSD functions)
  produce warning message for truncated course code, timetable, title

Version 2.4, 2006-04-12

  added undelete command
  no warnings using -Wall -Wextra
  allow for cut/paste of minis that are targets of arrow
  test for non-existence of source/target box/mini for new arrow
  handle more signals than SIGINT (such as SIGSEGV)

Version 2.3, 2006-03-23

  remove "which" checking of pdfviewer to allow, e.g., xpdf -remote label
  use geometry.sty for new blank TeX file
  declarations of optarg, optind moved to prerex.h

Version 2.2.2, 2006-03-08

  Free linked structures and readline strings.
  Turn grid off initially before restoring a backup.

Version 2.2.1, 2006-02-18

  Don't try to remove/restore access to an empty filename.

Version 2.2, 2006-02-15

  Inputs and outputs \begin{chart} ... \end{chart} instead of \chart{ ... }

Version 2.1.1, 2006-02-13

  Add designation PRIVATE ( = static) to definitions of functions and variables.

Version 2.1, 2006-02-09

  Trap SIGINT in order to restore write-access after ^C.
  Add Backup command.

Version 2.0.3, 2006-02-07

  New arrow should have default curvature.

Version 2.0.1, 2006-02-06

 Free list structures before Restoring.

Version 2.0, 2006-02-03

  Partition into separate modules prerex.c, inout.c, edit.c

Version 1.2.2, 2006-02-02

  Cleanup curvature code.
  Use %s to output % in regenerate_tex_file.

Version 1.2.1, 2006-02-02

  Allow for curved recommended and co-requisite arrows.

Version 1.2.0, 2006-02-01

  Restore command.

Version 1.1.2, 2006-01-31

  Create backup of Tex file before editing.
  Re-open TeX file for writing before regenerating.

Version 1.1.1, 2006-01-30

  TeX file read-only during editing (except when re-generating it).

Version 1.1.0, 2006-01-29

  Corrected use of cut stacks.
  Eliminated cut arrow command.
  Introduced delete box/mini and delete arrow.
  Checked for duplicate boxes, minis or arrows (e.g., when pasting).

Version 1.0.5, 2006-01-28

  Keep stacks of cut boxes, minis, and arrows.

Version 1.0.4, 2006-01-27

  Add option -p.

Version 1.0.3, 2006-01-26

  Replace system calls to "xpdf" by calls to pdfviewer (if defined).

Version 1.0.2, 2006-01-25

  Replace system call to "pdflatex filename" by call to "make filename.pdf"
    to provide more flexibility (e.g., LaTeX vs pdfLaTeX).

Version 1.0.1, 2006-01-24
  Error message for pdflatex failure now refers to filename.log.
  Inserted checks for lp or cp increasing too far in read_textfield 
    and read_bracketed_textfield.
  Process TeX file before analyzing it (as a syntax sanity pre-check).
  Defer arrow output if source box not previously output.

*/

# include "prerex.h"


extern int optind;              /* defined in unistd.h  */
char *pdfviewer; 
char *ps2pdf;

bool grid;			/* true if coordinate grid should be shown in background */
bool useLatex = false;          /* true if latex -> dvips -> ps2pdf should be used */
bool reprocess = true;          /* false if regenerate_and_process should do nothing */ 

char *command_line;

char basefilename[FILE_LEN];
char *basefilename_n = basefilename;
char chartfilename[FILE_LEN];
char *chartfilename_n = chartfilename;
char backup_filename[FILE_LEN + 1];
char *backup_filename_n = backup_filename;

FILE *tex_file, *backup_tex_file;

pid_t pid_pdfviewer = 0;                  /* process id for pdfviewer  */

void
usage (void)
{
  puts ("Usage: prerex [-l | -v | -h] [basefile[.tex] [chartfile[.tex]]]");
}

void
remove_write_access (void)	/* to prevent interference */
{
  char command[64] = {'\0'};
  char *command_n = command;
  append (command, &command_n, "chmod a-w ", sizeof (command));
  append (command, &command_n, chartfilename, sizeof (command));
  system (command);
}

void
restore_write_access (void)	/* to regenerate or restore the tex_file */
{
  char command[64] = {'\0'};
  char *command_n = command;
  if (chartfilename == NULL)
    return;
  append (command, &command_n, "chmod ug+w ", sizeof (command));
  append (command, &command_n, chartfilename, sizeof (command));
  system (command);
}

char deftext[LINE_LEN];		/* buffer for default input; set using set_deftext */
char *deftext_n = deftext;

int
set_deftext (void)		/* used by readline to output defaults */
{
  rl_insert_text (&deftext[0]);
  deftext[0] = '\0';
  deftext_n = deftext;
  rl_startup_hook = (rl_hook_func_t *) NULL;
  return 0;
}


PRIVATE void
sigproc (int i)  /* signal handler */
{
  i = i;	/* just to suppress a gcc warning message */
  grid = false;
  puts("Turning off coordinate grid.");
  regenerate_tex_file(); 
  fclose(tex_file);
  restore_write_access ();
  if (pid_pdfviewer > 0) kill(pid_pdfviewer, SIGTERM); /* terminate pdfviewer process */
  exit (0);
}


PRIVATE void
create_blank_tex_file (void)
{
  tex_file = fopen (chartfilename, "w+");
  if (tex_file == NULL)
    return;
  fprintf (tex_file, "%s\n", "\\documentclass{article}");
  fprintf (tex_file, "%s\n", "\\usepackage{geometry}");
  fprintf (tex_file, "%s\n", "\\geometry{noheadfoot, margin=0.5in}");
  fprintf (tex_file, "%s\n", "\\usepackage{prerex}");
  fprintf (tex_file, "%s\n", "\\begin{document}");
  fprintf (tex_file, "%s\n", "\\thispagestyle{empty}");
  fprintf (tex_file, "%s\n", "\\begin{chart}");
  fprintf (tex_file, "%s\n", "\\grid");
  fprintf (tex_file, "%s\n", "\\end{chart}");
  fprintf (tex_file, "%s\n", "\\end{document}");
  rewind (tex_file);
}


PRIVATE void
open_tex_file (void)
{
  char command[LINE_LEN] = {'\0'};
  char *command_n = command;
  tex_file = fopen (chartfilename, "r+");
  while (tex_file == NULL)
    {
      char code[16];
      deftext[0] = '\0';
      deftext_n = deftext;
      append (deftext, &deftext_n, "y", sizeof (deftext));
      rl_startup_hook = set_deftext;
      tex_file = fopen (chartfilename, "r");
      if (tex_file != NULL)
	{
	  command_line = readline ("This file is currently read-only; change permissions and continue (y/n)? ");
	  sscanf (command_line, "%15s", code);
	  free (command_line);
	  if (code[0] == 'y')
	    {
	      restore_write_access ();
	      tex_file = fopen (chartfilename, "r+");
	      continue;
	    }
	}
      command[0] = '\0';
      command_n = command;
      append (command, &command_n, "Create new prerex chart file ", sizeof (command));
      append (command, &command_n, chartfilename, sizeof (command));
      append (command, &command_n, " (y/n)? ", sizeof (command));
      command_line = readline (command);
      sscanf (command_line, "%15s", code);
      free (command_line);
      if (code[0] == 'y')
	{
	  create_blank_tex_file ();
	}
      else
	{
	  if (code[0] != 'n')
	    puts ("Response not recognized; 'n' assumed.");
          deftext[0] = '\0';
          deftext_n = deftext;
          append ( deftext, &deftext_n, "y", sizeof(deftext));
          rl_startup_hook = set_deftext;
          command_line = readline ("Quit (y/n)? ");
          sscanf (command_line, "%15s", code);
          free (command_line);
          if (code[0] == 'y') exit (0);
	  if (code[0] != 'n') puts ("Response not recognized; 'n' assumed.");
	  command_line = readline ("Enter another file name: ");
          basefilename[0] = '\0';
          basefilename_n = basefilename;
	  sscanf (command_line, "%127s", basefilename);
          while (*basefilename_n)  
            basefilename_n++;
	  free (command_line);
	  if (!suffix (".tex", basefilename))
	    append (basefilename, &basefilename_n, ".tex", sizeof (basefilename));
          chartfilename[0] = '\0';
          chartfilename_n = chartfilename;
          append (chartfilename, &chartfilename_n, basefilename, sizeof (chartfilename));
	  tex_file = fopen (chartfilename, "r+");
	}
    }
  remove_write_access ();
}

void
create_backup (void)
{
  backup_filename[0] = '\0';
  backup_filename_n = backup_filename;
  append (backup_filename, &backup_filename_n, ".", sizeof (backup_filename));
  append (backup_filename, &backup_filename_n, chartfilename, sizeof (backup_filename));
  backup_tex_file = fopen (backup_filename, "w+");
  if (backup_tex_file == NULL)
    {
      puts ("Can't open backup file.");
      return;
    }
  copy (tex_file, backup_tex_file);
  fclose (backup_tex_file);
}

bool
process_tex_file (void)
{ /*  generates system calls to make file.pdf */
  /*  returns true if processing fails */
  char *p;
  char command[LINE_LEN] = {'\0'};
  char *command_n = command;
  char filename_root[FILE_LEN + 3] = {'\0'};
  char *filename_root_n = filename_root;
  rewind (tex_file);
  append (filename_root, &filename_root_n, basefilename, sizeof (filename_root));
  p = strstr (filename_root, ".tex");
  *p = '\0';
  filename_root_n = p;
  if (useLatex)
    {
      printf ("Processing %s using LaTeX, dvips, and ps2pdf.\n", basefilename);
      append (command, &command_n, "latex ", sizeof (command));
    }
  else
    {
      printf ("Processing %s using pdfLaTeX.\n", basefilename);
      append (command, &command_n, "pdflatex ", sizeof (command));
    }
  append (command, &command_n,
           "-halt-on-error -interaction batchmode ", sizeof (command));
  append (command, &command_n, filename_root, sizeof (command));
  append (command, &command_n, " > /dev/null", sizeof (command));
  if (system (command))
    { /* try to display error message using PAGER or tail: */
      char *pager = getenv ("PAGER");
      command[0] = '\0';
      command_n = command;
      if (strcmp (pager, "more") == 0)
        append (command, &command_n, "more +/^! ", sizeof (command));
      else if (strcmp (pager, "less") == 0)
        append (command, &command_n, "less -p^! ", sizeof (command));
      else if (strcmp (pager, "most") == 0)
        append (command, &command_n, "most +/'! ' ", sizeof (command));
      else
        {
          puts ("    ...    [log file lines deleted] ");
          puts ("");
          append (command, &command_n, "tail -n 25 ", sizeof (command));
        }
      append (command, &command_n, filename_root, sizeof (command));
      append (command, &command_n, ".log", sizeof (command));
      system (command);
      return true;
    }
    if (useLatex)
      {
        command[0] = '\0';
        command_n = command;
        append (command, &command_n, "dvips -q ", sizeof (command));
        append (command, &command_n, filename_root, sizeof (command));
        append (command, &command_n, " -o ", sizeof (command));
        append (command, &command_n, filename_root, sizeof (command));
        append (command, &command_n, ".ps", sizeof (command));
        if (system (command))
        {
            puts ("dvips processing fails.");
            return true;
        }
        command[0] = '\0';
        command_n = command;
        append (command, &command_n, ps2pdf, sizeof (command));
        append (command, &command_n, filename_root, sizeof (command));
        append (command, &command_n, ".ps", sizeof (command));
        if (system (command))
        {
            puts ("ps2pdf processing fails.");
            return true;
        }
      }
    return false;
}

PRIVATE void
open_viewer (void)
{
  char *p;
  char pdf_filename[FILE_LEN];
  char *pdf_filename_n = pdf_filename;
  pdfviewer = getenv("PDFVIEWER");
  if (pdfviewer == NULL)
  {
    puts("PDFVIEWER is not defined in your environment.");
    return;
  }
  pdf_filename[0] = '\0';
  pdf_filename_n = pdf_filename;
  append (pdf_filename, &pdf_filename_n, basefilename, sizeof (pdf_filename));
  p = strstr (pdf_filename, ".tex");
  *p = '\0';
  pdf_filename_n = p;
  append (pdf_filename, &pdf_filename_n, ".pdf", sizeof (pdf_filename));
  printf ("Opening %s with %s.\n", pdf_filename, pdfviewer);
  pid_pdfviewer = fork();
  if (pid_pdfviewer == 0)  /* child process */
  {
     char *argv[3];
     argv[0] = pdfviewer;
     argv[1] = pdf_filename;
     argv[2] = (char *) NULL;
     execvp(argv[0], argv); 
     /* the call above returns only in case of an error */
     printf("%s fails to start.\n", pdfviewer); 
     _exit (0);
  }
  else if (pid_pdfviewer == -1) perror("fork failed");
  return;
}


void
regenerate_and_process (void)
{
  if (!reprocess) return;
  regenerate_tex_file ();
  process_tex_file ();  /* ignore failure; allows user to recover by editing the tex file */
}

int
main (int argc, char *argv[])
{
  int ch;
  const char version[] = "5.4, 2009-11-12";

  printf ("This is prerex, version %s.\n", version);
  ch = getopt (argc, argv, "hvl");
  while (ch != -1)
    {
      switch (ch)
	{
	case 'h':
	  usage ();
	  exit (0);
	case 'v':
	  exit (0);
        case 'l':
          useLatex = true;
          break;
	case '?':
	  exit (EXIT_FAILURE);
	default:
	  printf ("Function getopt returned character code 0%o.\n",
		  (unsigned int) ch);
	  exit (EXIT_FAILURE);
	}
      ch = getopt (argc, argv, "hvl");
    }
  puts ("Copyright (c) 2005-09  R. D. Tennent");
  puts ("School of Computing, Queen's University, rdt@cs.queensu.ca");
  puts
    ("This program comes with NO WARRANTY. You can redistribute prerex and/or");
  puts
    ("modify it under the terms of the GNU General Public License version 2.");
  puts ("");
  basefilename[0] = '\0';
  basefilename_n = basefilename;
  if (optind >= argc)
    {
      puts ("");
      do
	{
	  command_line = readline ("Please enter a file name: ");
	  sscanf (command_line, "%127s", basefilename);
          while (*basefilename_n)  
            basefilename_n++;
	  free (command_line);
	}
      while (basefilename[0] == '\0');
    }
  else
    append (basefilename, &basefilename_n, argv[optind], sizeof (basefilename));
  if (!suffix (".tex", basefilename))
    append (basefilename, &basefilename_n, ".tex", sizeof (basefilename));
  optind++;
  chartfilename[0] = '\0';
  chartfilename_n = chartfilename;
  if (optind < argc)  /* user-provided chartfilename */
    {
      append (chartfilename, &chartfilename_n, argv[optind], sizeof (chartfilename));
      if (!suffix (".tex", chartfilename))
        append (chartfilename, &chartfilename_n, ".tex", sizeof (chartfilename));
    }
  else /* use basefile as chartfile */
    append (chartfilename, &chartfilename_n, basefilename, sizeof (chartfilename));
  
  open_tex_file ();

  if (useLatex)
  {
     ps2pdf = getenv("PS2PDF");
     if (ps2pdf == NULL) ps2pdf = "ps2pdf";
     printf("Using %s for ps to pdf conversion.\n", ps2pdf);
  }

  if (process_tex_file ()) /* bail if initial processing fails */
  {
    restore_write_access ();
    exit(EXIT_FAILURE);
  }

  analyze_tex_file ();

  /* catch signals in order to regenerate tex file, restore write access, etc. */
  signal (SIGHUP, &sigproc);  
  signal (SIGINT, &sigproc); 
  signal (SIGQUIT, &sigproc);
  signal (SIGILL, &sigproc);
  signal (SIGSEGV, &sigproc);
  signal (SIGTERM, &sigproc);  

  grid = true;	
  puts("Turning on coordinate grid.");
  regenerate_and_process (); 

  open_viewer(); 

  create_backup ();  

  process_commands();

  return 0;
}
